/*
 * Copyright (c) 2016-2021 Satlab A/S <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

/**
 * @file polaris_boot.h
 *
 * This file contains functions to configure the boot process on a Polaris
 * system.
 *
 * The system uses two boot slots: slot 0 and slot 1. On system boot, a small
 * bootloader application selects which of the two slots to boot based on a
 * non-volatile state stored in FRAM.
 *
 * Each firmware slot has a set of properties:
 *
 * - status: either STABLE, TESTING, or DISABLED.
 * - attempts: remaining boot attempts.
 * - priority: higher priority number indicate higher preference to boot.
 *
 * The current bootloader status can be queried using @ref
 * sl_polaris_bootloader_get_status or shown with the `boot status` shell
 * command.
 *
 * A firmware slot is bootable if status is not DISABLED and attempts > 0. The
 * highest priority bootable slot is selected for boot.
 *
 * If no bootable slot is found, e.g. because all slots have attempts = 0,
 * attempts for all slots are reset to the default value (3) and another search
 * for a bootable slot is performed. If a bootable slot is still not found,
 * slot 0 is selected for boot.
 *
 * Before booting the selected slot, the bootloader decrements remaining
 * attempts for the slot. It is up to the booted firmware image to reset
 * remaining attempts to indicate a successful boot. For slots where status is
 * STABLE, the firmware will do so automatically after 120 seconds uptime
 * upon successful boot. For slots where status is TESTING, the attempts must
 * be reset using @ref sl_polaris_bootloader_boot_success or alternatively
 * using the `boot success` shell command.
 *
 * Each firmware slot can be flashed from the other using the @ref
 * sl_polaris_bootloader_flash function or the `boot flash` shell command.
 * During the firmware update, the slot being updated is automatically set to
 * DISABLED to prevent it from being booted. If the update is completed
 * successfully, the slot status is set to TESTING. The user can then change
 * the preferred slot to the updated slot using @ref
 * sl_polaris_bootloader_set_preferred or the `boot prefer` shell command. When
 * successfully booted, the user can decide to keep the slot in TESTING and
 * manually reset attempts as a safeguard, or mark the slot as STABLE to enable
 * automatic reset of remaining attempts.
 */

#ifndef _SL_POLARIS_CLIENT_BOOT_H_
#define _SL_POLARIS_CLIENT_BOOT_H_

#include <stdint.h>

/**
 * @brief Flash image file from SD-card to firmware slot
 *
 * This function flashes a firmware file from the SD-card to a firmware slot.
 *
 * Flashing and verifying a firmware image typically takes 15-30 seconds, so
 * ensure that the timeout is set sufficiently high.
 *
 * It is not possible to flash the currently booted firmware slot.
 *
 * @param node CSP address of the property node.
 * @param timeout Timeout of the command in milliseconds.
 * @param slot Firmware slot to flash (0 or 1).
 * @param filename Filename of the remote file to flash.
 *
 * @returns 0 on success, negative error code otherwise.
 */
int sl_polaris_bootloader_flash(uint8_t node, uint32_t timeout, uint8_t slot, const char *filename);

struct sl_polaris_bootloader_status {
	/** Booted slot */
	uint8_t slot_booted;
	/** Preferred slot */
	uint8_t slot_preferred;
	/** Slot 0 status */
	uint8_t slot0_status;
	/** Slot 0 priority */
	uint8_t slot0_attempts;
	/** Slot 0 remaining boot attempts */
	uint8_t slot0_priority;
	/** Slot 0 contents state */
	uint8_t slot0_valid;
	/** Slot 1 status */
	uint8_t slot1_status;
	/** Slot 1 priority */
	uint8_t slot1_attempts;
	/** Slot 1 remaining boot attempts */
	uint8_t slot1_priority;
	/** Slot 1 contents state */
	uint8_t slot1_valid;
};

/**
 * @brief Get bootloader status
 *
 * This function reads the current bootloader status.
 *
 * @param node CSP address of the property node.
 * @param timeout Timeout of the command in milliseconds.
 * @param status Status is returned in this structure.
 *
 * @returns 0 on success, negative error code otherwise.
 */
int sl_polaris_bootloader_get_status(uint8_t node, uint32_t timeout, struct sl_polaris_bootloader_status *status);

/**
 * @brief Mark boot successful
 *
 * This function marks the current boot successful and resets the boot attempts
 * for the booted slot to the default value. For testing slots, this function
 * needs to be called before attempts reaches 0, to keep the slot bootable.
 *
 * @param node CSP address of the property node.
 * @param timeout Timeout of the command in milliseconds.
 *
 * @returns 0 on success, negative error code otherwise.
 */
int sl_polaris_bootloader_boot_success(uint8_t node, uint32_t timeout);

/**
 * @brief Set the preferred boot slot
 *
 * This function is used to set the preferred slot to boot. It modifies the
 * priorities of all boot slots to ensure the selected slot has the highest. If
 * also reset the attempts of the preferred slot to the default value.
 *
 * @param node CSP address of the property node.
 * @param timeout Timeout of the command in milliseconds.
 * @param slot Firmware slot to prefer (0 or 1).
 *
 * @returns 0 on success, negative error code otherwise.
 */
int sl_polaris_bootloader_set_preferred(uint8_t node, uint32_t timeout, uint8_t slot);

/**
 * @brief Set slot status
 *
 * This function is used to set the status of a slot.
 *
 * To prevent deadlock, it is not possible to set the status of the booted slot
 * to DISABLED. Likewise, only the booted slot can be set to STABLE.
 *
 * @param node CSP address of the property node.
 * @param timeout Timeout of the command in milliseconds.
 * @param slot Firmware slot to change status of (0 or 1).
 * @param status New status for slot.
 *
 * @returns 0 on success, negative error code otherwise.
 */
int sl_polaris_bootloader_set_slot_status(uint8_t node, uint32_t timeout, uint8_t slot, uint8_t status);

/**
 * @brief Set slot priority
 *
 * This function is used to set the priority of a slot. Bootable slots with
 * higher priority are preferred.
 *
 * It is usually not necessary to call this function directly. Use @ref
 * sl_polaris_bootloader_set_preferred instead.
 *
 * @param node CSP address of the property node.
 * @param timeout Timeout of the command in milliseconds.
 * @param slot Firmware slot to change status of (0 or 1).
 * @param priority New priority for slot.
 *
 * @returns 0 on success, negative error code otherwise.
 */
int sl_polaris_bootloader_set_slot_priority(uint8_t node, uint32_t timeout, uint8_t slot, uint8_t priority);

/**
 * @brief Set slot attempts
 *
 * This function is used to set the remaining boot attempts of a slot.
 *
 * It is usually not necessary to call this function directly. Use @ref
 * sl_polaris_bootloader_set_preferred and @ref
 * sl_polaris_bootloader_boot_success instead.
 *
 * @param node CSP address of the property node.
 * @param timeout Timeout of the command in milliseconds.
 * @param slot Firmware slot to change status of (0 or 1).
 * @param attempts New attempts for slot.
 *
 * @returns 0 on success, negative error code otherwise.
 */
int sl_polaris_bootloader_set_slot_attempts(uint8_t node, uint32_t timeout, uint8_t slot, uint8_t attempts);

#endif /* _SL_POLARIS_CLIENT_BOOT_H_ */
