/*
 * Copyright (c) 2016-2021 Satlab A/S <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

/**
 * @file polaris_proto.h
 *
 * This file specifies the available CSP services and ports on the Polaris
 * receiver. It should not be necessary for clients to use these definitions
 * directly. Instead, use the wrapper functions from this library which handles
 * the protocol and performs necessary byte order conversion.
 *
 *
 */

#ifndef _SL_POLARIS_CLIENT_PROTO_H_
#define _SL_POLARIS_CLIENT_PROTO_H_

#include <stdint.h>

/** Default CSP address */
#define SL_POLARIS_DEFAULT_ADDRESS		20

/** CSP ports */
#define SL_POLARIS_PORT_SAMPLEBUF		20
#define SL_POLARIS_PORT_STORE			21
#define SL_POLARIS_PORT_BOOTLOADER		25 /* Deprecated */
#define SL_POLARIS_PORT_SHELL			26
#define SL_POLARIS_PORT_BOOTLOADER2		27

/** Default mirror frame receive port */
#define SL_POLARIS_PORT_STORE_MIRROR_RECV	30

/** Error codes */
#define SL_POLARIS_ERR_NONE			0
#define SL_POLARIS_ERR_INVAL			1
#define SL_POLARIS_ERR_IO			2
#define SL_POLARIS_ERR_NOMEM			3
#define SL_POLARIS_ERR_BUSY			4

/* Generic packet types */
typedef struct {
	uint8_t type;
} __attribute__ ((packed)) sl_polaris_req_t;

typedef struct {
	uint8_t type;
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_rep_t;

/*
 * SL_POLARIS_PORT_SAMPLEBUF
 */

/* Types */
#define SL_POLARIS_SAMPLEBUF_READ_REQ		1
#define SL_POLARIS_SAMPLEBUF_READ_REP		2
#define SL_POLARIS_SAMPLEBUF_READ_CHUNK		3
#define SL_POLARIS_SAMPLEBUF_FREE		4

/* Sources */
#define SL_POLARIS_SAMPLEBUF_SOURCE_RAW		0
#define SL_POLARIS_SAMPLEBUF_SOURCE_CH1		1
#define SL_POLARIS_SAMPLEBUF_SOURCE_CH2		2
#define SL_POLARIS_SAMPLEBUF_SOURCE_CH3		3
#define SL_POLARIS_SAMPLEBUF_SOURCE_CH4		4

#define SL_POLARIS_SAMPLEBUF_SOURCE_AIS1_CH87B  SL_POLARIS_SAMPLEBUF_SOURCE_CH3
#define SL_POLARIS_SAMPLEBUF_SOURCE_AIS2_CH88B  SL_POLARIS_SAMPLEBUF_SOURCE_CH4
#define SL_POLARIS_SAMPLEBUF_SOURCE_AIS3_CH75   SL_POLARIS_SAMPLEBUF_SOURCE_CH1
#define SL_POLARIS_SAMPLEBUF_SOURCE_AIS4_CH76   SL_POLARIS_SAMPLEBUF_SOURCE_CH2

/* Destinations */
#define SL_POLARIS_SAMPLEBUF_DEST_CSP		0
#define SL_POLARIS_SAMPLEBUF_DEST_MEM		1
#define SL_POLARIS_SAMPLEBUF_DEST_SDCARD	2

/* Flags */
#define SL_POLARIS_SAMPLEBUF_FL_FORCE_INTRAM	(1 << 0)
#define SL_POLARIS_SAMPLEBUF_FL_SYNC_PPS	(1 << 1)

typedef struct {
	uint8_t type;
	uint8_t source;
	uint8_t dest;
	uint8_t flags;
	uint32_t id;
	uint32_t samples;
} __attribute__ ((packed)) sl_polaris_samplebuf_read_req_t;

typedef struct {
	uint8_t type;
	uint8_t error;
	uint32_t id;
	uint32_t addr;
} __attribute__ ((packed)) sl_polaris_samplebuf_read_rep_t;

#define SL_POLARIS_SAMPLEBUF_READ_CHUNK_SAMPLES	32
typedef struct {
	uint8_t type;
	uint8_t error;
	uint16_t seq;
	uint16_t total;
	uint32_t data[SL_POLARIS_SAMPLEBUF_READ_CHUNK_SAMPLES];
} __attribute__ ((packed)) sl_polaris_samplebuf_read_chunk_t;

typedef struct {
	uint8_t type;
	uint8_t flags;
	uint32_t addr;
} __attribute__ ((packed)) sl_polaris_samplebuf_free_req_t;

typedef struct {
	uint8_t type;
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_samplebuf_free_rep_t;

/*
 * SL_POLARIS_PORT_STORE
 */

/* Types */
#define SL_POLARIS_STORE_FRAME			1 /* Deprecated */
#define SL_POLARIS_STORE_MIRROR_REQ		2
#define SL_POLARIS_STORE_MIRROR_REP		3
#define SL_POLARIS_STORE_READ_REQ		4
#define SL_POLARIS_STORE_READ_REP		5
#define SL_POLARIS_STORE_FRAME2 		6

#define SL_POLARIS_STORE_CHANNEL_MASK 		0x03
#define SL_POLARIS_STORE_FLAGS_MASK 		0xE0
#define SL_POLARIS_STORE_FLAGS_OFFSET 		5

#define SL_POLARIS_STORE_MAX_DATALEN		160
typedef struct {
	/** Must be SL_POLARIS_STORE_FRAME2 */
	uint8_t type;
	/** Error code */
	uint8_t error;
	/** Sequence number of this frame */
	uint32_t seq;
	/** Time of reception (nanoseconds since 1/1/1970) */
	uint64_t time;
	/** AIS channel this message was received on and pps flags.
	 *  bit 0-1: channel from 0 to 3
	 *  bit 2-4: reserved
	 *  bit 5: pps synchronized samples
	 *  bit 6: pps time set
	 *  bit 7: pps valid */
	uint8_t channel_and_flags;
	/** Signal power in 1/10 dBm */
	int16_t power_deci_dbm;
	/** Estimated frequency offset */
	int16_t frequency_hz;
	/** Number of bytes in data field */
	uint8_t length;
	/** Raw AIS message, excludes flag bytes, includes FCS */
	uint8_t data[SL_POLARIS_STORE_MAX_DATALEN];
} __attribute__ ((packed)) sl_polaris_store_frame_t;
#define SL_POLARIS_STORE_FRAME_OVERHEAD		(sizeof(sl_polaris_store_frame_t) - SL_POLARIS_STORE_MAX_DATALEN)

typedef struct {
	uint8_t type;
	uint8_t dest_node;
	uint8_t dest_port;
	uint32_t timeout_ms;
} __attribute__ ((packed)) sl_polaris_store_mirror_req_t;

typedef struct {
	uint8_t type;
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_store_mirror_rep_t;

/* Bitmap size */
#define SL_POLARIS_STORE_BITMAP_SIZE		32 /* bytes */

typedef struct {
	uint8_t type;
	uint32_t start;
	uint8_t bitmap[SL_POLARIS_STORE_BITMAP_SIZE];
} __attribute__ ((packed)) sl_polaris_store_read_req_t;

typedef struct {
	uint8_t type;
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_store_read_rep_t;

/*
 * SL_POLARIS_PORT_SHELL
 */

/* Types */
#define SL_POLARIS_SHELL_RUN_REQ		1
#define SL_POLARIS_SHELL_RUN_REP		2

/* Keys */
#define SL_POLARIS_SHELL_RUN_KEY		0x3fb95c30

#define SL_POLARIS_SHELL_RUN_MAX		100
typedef struct {
	/** Must be SL_POLARIS_SHELL_RUN_REQ */
	uint8_t type;
	/** Flags, currently unused - set to zero */
	uint8_t flags;
	/** Must be SL_POLARIS_SHELL_RUN_KEY */
	uint32_t key;
	/** Command to execute */
	uint8_t cmd[SL_POLARIS_SHELL_RUN_MAX];
} __attribute__ ((packed)) sl_polaris_shell_run_req_t;

typedef struct {
	/** Must be SL_POLARIS_SHELL_RUN_REP */
	uint8_t type;
	/** Error code */
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_shell_run_rep_t;

/*
 * SL_POLARIS_PORT_BOOTLOADER2
 */

/* Types */
#define SL_POLARIS_BOOTLOADER2_STATUS_REQ		1
#define SL_POLARIS_BOOTLOADER2_STATUS_REP		2
#define SL_POLARIS_BOOTLOADER2_BOOT_SUCCESS_REQ		3
#define SL_POLARIS_BOOTLOADER2_BOOT_SUCCESS_REP		4
#define SL_POLARIS_BOOTLOADER2_SET_PREFERRED_REQ	5
#define SL_POLARIS_BOOTLOADER2_SET_PREFERRED_REP	6
#define SL_POLARIS_BOOTLOADER2_FLASH_REQ		10
#define SL_POLARIS_BOOTLOADER2_FLASH_REP		11
#define SL_POLARIS_BOOTLOADER2_SET_SLOT_STATUS_REQ	20
#define SL_POLARIS_BOOTLOADER2_SET_SLOT_STATUS_REP	21
#define SL_POLARIS_BOOTLOADER2_SET_SLOT_PRIORITY_REQ	22
#define SL_POLARIS_BOOTLOADER2_SET_SLOT_PRIORITY_REP	23
#define SL_POLARIS_BOOTLOADER2_SET_SLOT_ATTEMPTS_REQ	24
#define SL_POLARIS_BOOTLOADER2_SET_SLOT_ATTEMPTS_REP	25

/* Keys */
#define SL_POLARIS_BOOTLOADER2_BOOT_SUCCESS_KEY		0x0f9de49f
#define SL_POLARIS_BOOTLOADER2_SET_PREFERRED_KEY	0x61b6f00e
#define SL_POLARIS_BOOTLOADER2_FLASH_KEY		0xc87042c3
#define SL_POLARIS_BOOTLOADER2_SET_SLOT_STATUS_KEY	0x512a753a
#define SL_POLARIS_BOOTLOADER2_SET_SLOT_PRIORITY_KEY	0xcc836cf8
#define SL_POLARIS_BOOTLOADER2_SET_SLOT_ATTEMPTS_KEY	0xadb5064a

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_STATUS_REQ */
	uint8_t type;
	/* Flags, currently unused - set to zero */
	uint8_t flags;
} __attribute__ ((packed)) sl_polaris_bootloader2_status_req_t;

#define SL_POLARIS_BOOTLOADER2_STATUS_SLOT_DISABLED	0
#define SL_POLARIS_BOOTLOADER2_STATUS_SLOT_TESTING	1
#define SL_POLARIS_BOOTLOADER2_STATUS_SLOT_STABLE	2

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_STATUS_REP */
	uint8_t type;
	/* Error code */
	uint8_t error;
	/* Booted slot */
	uint8_t slot_booted;
	/* Preferred slot */
	uint8_t slot_preferred;
	/* Slot 0 status */
	uint8_t slot0_status;
	/* Slot 0 priority */
	uint8_t slot0_attempts;
	/* Slot 0 remaining boot attempts */
	uint8_t slot0_priority;
	/* Slot 0 contents state */
	uint8_t slot0_valid;
	/* Slot 1 status */
	uint8_t slot1_status;
	/* Slot 1 priority */
	uint8_t slot1_attempts;
	/* Slot 1 remaining boot attempts */
	uint8_t slot1_priority;
	/* Slot 1 contents state */
	uint8_t slot1_valid;
} __attribute__ ((packed)) sl_polaris_bootloader2_status_rep_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_BOOT_SUCCESS_REQ */
	uint8_t type;
	/* Must be SL_POLARIS_BOOTLOADER2_BOOT_SUCCESS_KEY */
	uint32_t key;
} __attribute__ ((packed)) sl_polaris_bootloader2_boot_success_req_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_BOOT_SUCCESS_REP */
	uint8_t type;
	/* Error code */
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_bootloader2_boot_success_rep_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_SET_PREFERRED_REQ */
	uint8_t type;
	/* Must be SL_POLARIS_BOOTLOADER2_SET_PREFERRED_KEY */
	uint32_t key;
	/* Preferred slot */
	uint8_t slot;
} __attribute__ ((packed)) sl_polaris_bootloader2_set_preferred_req_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_SET_PREFERRED_REP */
	uint8_t type;
	/* Error code */
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_bootloader2_set_preferred_rep_t;

#define SL_POLARIS_BOOTLOADER2_FILENAME_MAX	40
typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_FLASH_REQ */
	uint8_t type;
	/* Flags, currently unused - set to zero */
	uint8_t flags;
	/* Must be SL_POLARIS_BOOTLOADER2_FLASH_KEY */
	uint32_t key;
	/* Slot to flash */
	uint8_t slot;
	/* Filename to flash to the alternate partition */
	uint8_t filename[SL_POLARIS_BOOTLOADER2_FILENAME_MAX];
} __attribute__ ((packed)) sl_polaris_bootloader2_flash_req_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_FLASH_REP */
	uint8_t type;
	/* Error code */
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_bootloader2_flash_rep_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_SET_SLOT_STATUS_REQ */
	uint8_t type;
	/* Must be SL_POLARIS_BOOTLOADER2_SET_SLOT_STATUS_KEY */
	uint32_t key;
	/* Slot to change status of */
	uint8_t slot;
	/* New status */
	uint8_t status;
} __attribute__ ((packed)) sl_polaris_bootloader2_set_slot_status_req_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_SET_SLOT_STATUS_REP */
	uint8_t type;
	/* Error code */
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_bootloader2_set_slot_status_rep_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_SET_SLOT_PRIORITY_REQ */
	uint8_t type;
	/* Must be SL_POLARIS_BOOTLOADER2_SET_SLOT_PRIORITY_KEY */
	uint32_t key;
	/* Slot to change priority of */
	uint8_t slot;
	/* New priority */
	uint8_t priority;
} __attribute__ ((packed)) sl_polaris_bootloader2_set_slot_priority_req_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_SET_SLOT_PRIORITY_REP */
	uint8_t type;
	/* Error code */
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_bootloader2_set_slot_priority_rep_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_SET_SLOT_ATTEMPTS_REQ */
	uint8_t type;
	/* Must be SL_POLARIS_BOOTLOADER2_SET_SLOT_ATTEMPTS_KEY */
	uint32_t key;
	/* Slot to change attempts of */
	uint8_t slot;
	/* New attempts */
	uint8_t attempts;
} __attribute__ ((packed)) sl_polaris_bootloader2_set_slot_attempts_req_t;

typedef struct {
	/* Must be SL_POLARIS_BOOTLOADER2_SET_SLOT_ATTEMPTS_REP */
	uint8_t type;
	/* Error code */
	uint8_t error;
} __attribute__ ((packed)) sl_polaris_bootloader2_set_slot_attempts_rep_t;

#endif /* _SL_POLARIS_CLIENT_PROTO_H_ */
