from threading import Thread
import queue
from datetime import datetime

import satlab.polaris_ais_client as ais_client

class MirrorTask:
    """ Helper class for receiving mirror frames from Polaris-AIS

    This class can be used as a context manager or alternativly the `start` and `stop`
    methods can be used. A queue is created for holding mirrored AIS frames and they
    can be retrived with the `get` method.
    """

    def __init__(self, client):
        """Init MirrorTask

        Args:
            client: polaris_ais_client instance
        """
        self._client = client
        self._queue = queue.Queue()
        self._thread = None
        self._thread_stop = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stop()

    def start(self):
        """Request receiver to start mirroring frames

        This method spawn a new thread to run the mirror loop.
        When done call `stop` to both tell the receiver to stop mirroring frames
        and to stop remove the thread spawned.
        """
        self._thread_stop = False
        self._thread = Thread(target=self._store_mirror_thread)
        self._thread.start()

    def stop(self):
        """Request receiver to stop mirroring frames

        A `RuntimeError` exception is raised if `stop` is called before `start`.
        """
        self._thread_stop = True
        if self._thread is None:
            raise RuntimeError("Call start before stop")
        self._thread.join()
        self._thread = None

    # Override get to catch exception from mirror thread that is also placed on the queue
    # together with frames
    def get(self, block=True, timeout=None):
        """Remove and return an item from the AIS frame queue

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).
        """
        res = self._queue.get(block=block, timeout=timeout)
        if not isinstance(res, ais_client.PolarisFrame):
            raise res
        return res

    def _store_mirror_thread(self):
        while not self._thread_stop:
            try:
                self._client.store_mirror_recv(self._store_mirror_cb)
            except Exception as e:
                # Put exception in queue so get call can reraise it
                self._queue.put(e)
                break

    def _store_mirror_cb(self, frame):
        # Can be called with None so check if it is a frame before enqueuing
        if frame:
            self._queue.put(frame)

        if self._thread_stop:
            return -1

        return 0
