/*
 * Copyright (c) 2016-2021 Satlab A/S <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

#include <satlab/polaris_boot.h>

#include <errno.h>

#include <csp/csp.h>
#include <csp/csp_endian.h>

#include <satlab/polaris.h>

int sl_polaris_bootloader_flash(uint8_t node, uint32_t timeout, uint8_t slot, const char *filename)
{
	int ret;
	sl_polaris_bootloader2_flash_req_t request;
	sl_polaris_bootloader2_flash_rep_t reply;

	request.type = SL_POLARIS_BOOTLOADER2_FLASH_REQ;
	request.flags = 0;
	request.slot = slot;
	request.key = csp_hton32(SL_POLARIS_BOOTLOADER2_FLASH_KEY);

	if (strlen(filename) >= sizeof(request.filename))
		return -ENOSPC;

	strncpy((char *)request.filename, filename, sizeof(request.filename));
	request.filename[sizeof(request.filename) - 1] = '\0';

	ret = csp_transaction(CSP_PRIO_NORM, node, SL_POLARIS_PORT_BOOTLOADER2, timeout,
			      &request, sizeof(request), &reply, sizeof(reply));
	if (ret != sizeof(reply))
		return -ECONNREFUSED;

	if (reply.type != SL_POLARIS_BOOTLOADER2_FLASH_REP)
		return -EINVAL;

	if (reply.error != SL_POLARIS_ERR_NONE)
		return -reply.error;

	return 0;
}

int sl_polaris_bootloader_get_status(uint8_t node, uint32_t timeout, struct sl_polaris_bootloader_status *status)
{
	int ret;
	sl_polaris_bootloader2_status_req_t request;
	sl_polaris_bootloader2_status_rep_t reply;

	if (!status)
		return -EINVAL;

	request.type = SL_POLARIS_BOOTLOADER2_STATUS_REQ;
	request.flags = 0;

	ret = csp_transaction(CSP_PRIO_NORM, node, SL_POLARIS_PORT_BOOTLOADER2, timeout,
			      &request, sizeof(request), &reply, sizeof(reply));
	if (ret != sizeof(reply))
		return -ECONNREFUSED;

	if (reply.type != SL_POLARIS_BOOTLOADER2_STATUS_REP)
		return -EINVAL;

	if (reply.error != SL_POLARIS_ERR_NONE)
		return -reply.error;

	status->slot_booted = reply.slot_booted;
	status->slot_preferred = reply.slot_preferred;
	status->slot0_status = reply.slot0_status;
	status->slot0_attempts = reply.slot0_attempts;
	status->slot0_priority = reply.slot0_priority;
	status->slot0_valid = reply.slot0_valid;
	status->slot1_status = reply.slot1_status;
	status->slot1_attempts = reply.slot1_attempts;
	status->slot1_priority = reply.slot1_priority;
	status->slot1_valid = reply.slot1_valid;

	return 0;
}

int sl_polaris_bootloader_boot_success(uint8_t node, uint32_t timeout)
{
	int ret;
	sl_polaris_bootloader2_boot_success_req_t request;
	sl_polaris_bootloader2_boot_success_rep_t reply;

	request.type = SL_POLARIS_BOOTLOADER2_BOOT_SUCCESS_REQ;
	request.key = csp_hton32(SL_POLARIS_BOOTLOADER2_BOOT_SUCCESS_KEY);

	ret = csp_transaction(CSP_PRIO_NORM, node, SL_POLARIS_PORT_BOOTLOADER2, timeout,
			      &request, sizeof(request), &reply, sizeof(reply));
	if (ret != sizeof(reply))
		return -ECONNREFUSED;

	if (reply.type != SL_POLARIS_BOOTLOADER2_BOOT_SUCCESS_REP)
		return -EINVAL;

	if (reply.error != SL_POLARIS_ERR_NONE)
		return -reply.error;

	return 0;
}

int sl_polaris_bootloader_set_preferred(uint8_t node, uint32_t timeout, uint8_t slot)
{
	int ret;
	sl_polaris_bootloader2_set_preferred_req_t request;
	sl_polaris_bootloader2_set_preferred_rep_t reply;

	request.type = SL_POLARIS_BOOTLOADER2_SET_PREFERRED_REQ;
	request.key = csp_hton32(SL_POLARIS_BOOTLOADER2_SET_PREFERRED_KEY);
	request.slot = slot;

	ret = csp_transaction(CSP_PRIO_NORM, node, SL_POLARIS_PORT_BOOTLOADER2, timeout,
			      &request, sizeof(request), &reply, sizeof(reply));
	if (ret != sizeof(reply))
		return -ECONNREFUSED;

	if (reply.type != SL_POLARIS_BOOTLOADER2_SET_PREFERRED_REP)
		return -EINVAL;

	if (reply.error != SL_POLARIS_ERR_NONE)
		return -reply.error;

	return 0;
}

int sl_polaris_bootloader_set_slot_status(uint8_t node, uint32_t timeout, uint8_t slot, uint8_t status)
{
	int ret;
	sl_polaris_bootloader2_set_slot_status_req_t request;
	sl_polaris_bootloader2_set_slot_status_rep_t reply;

	if (status != SL_POLARIS_BOOTLOADER2_STATUS_SLOT_DISABLED &&
	    status != SL_POLARIS_BOOTLOADER2_STATUS_SLOT_TESTING &&
	    status != SL_POLARIS_BOOTLOADER2_STATUS_SLOT_STABLE)
		return -EINVAL;

	request.type = SL_POLARIS_BOOTLOADER2_SET_SLOT_STATUS_REQ;
	request.key = csp_hton32(SL_POLARIS_BOOTLOADER2_SET_SLOT_STATUS_KEY);
	request.slot = slot;
	request.status = status;

	ret = csp_transaction(CSP_PRIO_NORM, node, SL_POLARIS_PORT_BOOTLOADER2, timeout,
			      &request, sizeof(request), &reply, sizeof(reply));
	if (ret != sizeof(reply))
		return -ECONNREFUSED;

	if (reply.type != SL_POLARIS_BOOTLOADER2_SET_SLOT_STATUS_REP)
		return -EINVAL;

	if (reply.error != SL_POLARIS_ERR_NONE)
		return -reply.error;

	return 0;
}

int sl_polaris_bootloader_set_slot_priority(uint8_t node, uint32_t timeout, uint8_t slot, uint8_t priority)
{
	int ret;
	sl_polaris_bootloader2_set_slot_priority_req_t request;
	sl_polaris_bootloader2_set_slot_priority_rep_t reply;

	request.type = SL_POLARIS_BOOTLOADER2_SET_SLOT_PRIORITY_REQ;
	request.key = csp_hton32(SL_POLARIS_BOOTLOADER2_SET_SLOT_PRIORITY_KEY);
	request.slot = slot;
	request.priority = priority;

	ret = csp_transaction(CSP_PRIO_NORM, node, SL_POLARIS_PORT_BOOTLOADER2, timeout,
			      &request, sizeof(request), &reply, sizeof(reply));
	if (ret != sizeof(reply))
		return -ECONNREFUSED;

	if (reply.type != SL_POLARIS_BOOTLOADER2_SET_SLOT_PRIORITY_REP)
		return -EINVAL;

	if (reply.error != SL_POLARIS_ERR_NONE)
		return -reply.error;

	return 0;
}

int sl_polaris_bootloader_set_slot_attempts(uint8_t node, uint32_t timeout, uint8_t slot, uint8_t attempts)
{
	int ret;
	sl_polaris_bootloader2_set_slot_attempts_req_t request;
	sl_polaris_bootloader2_set_slot_attempts_rep_t reply;

	request.type = SL_POLARIS_BOOTLOADER2_SET_SLOT_ATTEMPTS_REQ;
	request.key = csp_hton32(SL_POLARIS_BOOTLOADER2_SET_SLOT_ATTEMPTS_KEY);
	request.slot = slot;
	request.attempts = attempts;

	ret = csp_transaction(CSP_PRIO_NORM, node, SL_POLARIS_PORT_BOOTLOADER2, timeout,
			      &request, sizeof(request), &reply, sizeof(reply));
	if (ret != sizeof(reply))
		return -ECONNREFUSED;

	if (reply.type != SL_POLARIS_BOOTLOADER2_SET_SLOT_ATTEMPTS_REP)
		return -EINVAL;

	if (reply.error != SL_POLARIS_ERR_NONE)
		return -reply.error;

	return 0;
}
